from __future__ import absolute_import

__version__ = (0, 1, 12)  # remember to change setup.py

import contextlib
from functools import wraps
from threading import local
from itertools import cycle
from random import randint
from warnings import warn

from django.conf import settings
from django.utils import importlib

from .exceptions import PinDbException, PinDbConfigError, UnpinnedWriteException

__all__ = (
    'PinDbException', 'PinDbConfigError', 'UnpinnedWriteException',
    'unpin_all', 'pin', 'get_pinned', 'get_newly_pinned',
    'is_pinned', 'get_replica', 'unpinned_replica',
    'populate_replicas', 'StrictPinDbRouter', 'GreedyPinDbRouter'
)

_locals = local()

def is_enabled():
    return getattr(settings, 'PINDB_ENABLED', True)

def unpin_all():
    """Clear the new and old pinnings and the chosen replicas."""
    # the authoritative set of pinned aliases:
    _locals.pinned_set = set()
    # the newly-pinned ones for advising the pinned context (i.e. for persistence):
    _locals.newly_pinned_set = set()
    # replica choices already made during this pinning context:
    _locals.chosen_replicas = {}  # {master alias: replica alias}

# Number of replicas for each DB set, loaded when the Router is constructed;
# zero-based to ease using random.randint. If a set as 3 replicas, there will
# be a 2 here.
DB_SET_SIZES = {}  # How many slaves each DB set has - 1
def _init_state():
    if getattr(_locals, 'inited', False):
        return
    unpin_all()
    _locals.inited = True

def pin(alias, count_as_new=True):
    _init_state()
    _locals.pinned_set.add(alias)
    if count_as_new:
        _locals.newly_pinned_set.add(alias)

def _unpin_one(alias, also_unpin_new=True):
    """
    Not intended for external use; just here for the decorators below.
    """
    _init_state()
    _locals.pinned_set.remove(alias)
    if also_unpin_new:
        _locals.newly_pinned_set.discard(alias)

def get_pinned():
    _init_state()
    return _locals.pinned_set.copy()

def get_newly_pinned():
    _init_state()
    return _locals.newly_pinned_set.copy()

def is_pinned(alias):
    _init_state()
    return alias in _locals.pinned_set

def is_newly_pinned(alias):
    _init_state()
    return alias in _locals.newly_pinned_set

REPLICA_TEMPLATE = "%s-%s"
def _make_replica_alias(master_alias, replica_num):
    return REPLICA_TEMPLATE % (master_alias, replica_num)

def get_replica(master_alias):
    """Return an arbitrary replica of a given master.

    If one was already chosen during this pinning context, keep returning the
    same one.

    """
    _init_state()
    try:
        effective_size = DB_SET_SIZES[master_alias]
    except KeyError:
        # this happens if the main router isn't a PinDB router.
        #  in that case, we're meant to be disabled;
        #  just return master_alias as it's the best we can do.
        return master_alias
    if effective_size == -1:
        return master_alias
    else:
        previous_replica = _locals.chosen_replicas.get(master_alias)
        if previous_replica:
            return previous_replica
        replica_num = randint(0, effective_size)

        chosen_replica = _make_replica_alias(master_alias, replica_num)
        _locals.chosen_replicas[master_alias] = chosen_replica

        return chosen_replica

class unpinned_replica(object):
    """
    with unpinned_replica("default"):
        ...

    Read from a replica despite pinning state.
    """
    def __init__(self, alias):
        self.alias = alias

    def __enter__(self):
        self.was_pinned = is_pinned(self.alias)
        self.was_newly_pinned = is_newly_pinned(self.alias)
        if self.was_pinned:
            _unpin_one(self.alias, True)

    def __exit__(self, type, value, tb):
        if self.was_pinned:
            pin(self.alias, self.was_newly_pinned)

        if any((type, value, tb)):
            raise type, value, tb

def _mash_aliases(aliases):
    if not isinstance(aliases, basestring) and hasattr(aliases, '__iter__'):
        aliases = set(aliases)
    else:
        aliases = set([aliases])
    return aliases

def with_replicas(aliases):
    """
    @with_replicas([alias,...])
    def func...

    Read from replicas despite pinning state.
    """
    aliases = _mash_aliases(aliases)
    # FIXME: test this.
    def make_wrapper(func):
        replicas = [unpinned_replica(alias) for alias in aliases]
        @wraps(func)
        def wrapper(*args, **kwargs):
            with contextlib.nested(*replicas):
                return func(*args, **kwargs)
        return wrapper
    return make_wrapper

class master(object):
    """Context manager for temporarily writing to a master DB

    While active, any writes to the given DB set will go to the set's master,
    regardless of pinning state. Pinning state is not changed by this. ::

        with master("default"):
            ...

    """
    # TODO: make this optionally take a list of models for which to pin appropriately.
    def __init__(self, alias):
        self.alias = alias
        _init_state()

    def __enter__(self):
        self.was_pinned = is_pinned(self.alias)
        self.was_newly_pinned = is_newly_pinned(self.alias)
        pin(self.alias, False)

    def __exit__(self, type, value, tb):
        if not self.was_pinned:
            _unpin_one(self.alias, False)
        if not self.was_newly_pinned:
            # FIXME: should be abstracted somewhere, whoops.
            _locals.newly_pinned_set.discard(self.alias)

        if any((type, value, tb)):
            raise type, value, tb

def with_masters(aliases):
    """
    @with_masters([alias,...])
    def func...

    Write to masters despite (and without affecting) pinning state.
    """
    aliases = _mash_aliases(aliases)
    # FIXME: test this.
    def make_wrapper(func):
        masters = [master(alias) for alias in aliases]
        @wraps(func)
        def wrapper(*args, **kwargs):
            with contextlib.nested(*masters):
                return func(*args, **kwargs)
        return wrapper
    return make_wrapper

# TODO: add logging to aid debugging client code.
def populate_replicas(masters, replicas_overrides, unmanaged_default=False):
    if not 'default' in masters and not unmanaged_default:
        raise PinDbConfigError("You must declare a default master.")

    ret = {}
    for alias, master_values in masters.items():
        ret[alias] = master_values
        try:
            replica_overrides = replicas_overrides[alias]
        except KeyError:
            raise PinDbConfigError("No replica settings found for DB set %s" % alias)
        for i, replica_override in enumerate(replica_overrides):
            replica_alias = _make_replica_alias(alias, i)
            replica_settings = master_values.copy()
            replica_settings.update(replica_override)
            replica_settings['TEST_MIRROR'] = alias
            ret[replica_alias] = replica_settings

    return ret

class DummyRouter(object):
    def db_for_read(self, model, **hints):
        return "default"

    def db_for_write(self, model, **hints):
        return "default"

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_syncdb(slef, db, model):
        return True

class PinDbRouterBase(object):
    def __init__(self):
        if (not hasattr(settings, 'MASTER_DATABASES') or
            not hasattr(settings, 'DATABASE_SETS')):
            raise PinDbConfigError("You must define MASTER_DATABASES and DATABASE_SETS settings.")

        # stash the # to chose from to reduce per-call overhead in the routing.
        for alias, master_values in settings.MASTER_DATABASES.items():
            DB_SET_SIZES[alias] = len(settings.DATABASE_SETS[alias]) - 1
            if DB_SET_SIZES[alias] == -1:
                warn("No replicas found for %s; using just the master" % alias)

        # defer master selection to a domain-specific router.
        delegates = getattr(settings, 'PINDB_DELEGATE_ROUTERS', [])
        if delegates:
            from django.db.utils import ConnectionRouter
            self.delegate = ConnectionRouter(delegates)
        else:
            warn("Unable to load delegate router from settings.PINDB_DELEGATE_ROUTERS; using default and its replicas")
            # or just always use default's set.
            self.delegate = DummyRouter()

    def db_for_read(self, model, **hints):
        master_alias = self.delegate.db_for_read(model, **hints)
        if master_alias is None:
            master_alias = "default"

        if not is_enabled():
            return master_alias

        # allow anything unmanaged by the DB set system to work unhindered.
        if not master_alias in settings.MASTER_DATABASES:
            return master_alias

        if is_pinned(master_alias):
            return master_alias
        return get_replica(master_alias)

    def db_for_write(self, model, **hints):
        master_alias = self.delegate.db_for_write(model, **hints)
        if master_alias is None:
            master_alias = "default"

        if not is_enabled():
            return master_alias

        # allow anything unmanaged by the DB set system to work unhindered.
        if not master_alias in settings.MASTER_DATABASES:
            return master_alias
        return self._for_write_with_policy(master_alias, model, **hints)

    def allow_relation(self, obj1, obj2, **hints):
        return self.delegate.allow_relation(obj1, obj2, **hints)

    def allow_syncdb(self, db, model):
        return self.delegate.allow_syncdb(db, model)

class StrictPinDbRouter(PinDbRouterBase):
    def _for_write_with_policy(self, master_alias, model, **hints):
        if not is_pinned(master_alias):
            raise UnpinnedWriteException("Writes to %s aren't allowed because reads aren't pinned to it." % master_alias)
        return master_alias

class GreedyPinDbRouter(PinDbRouterBase):
    def _for_write_with_policy(self, master_alias, model, **hints):
        pin(master_alias)
        return master_alias



from django.test.simple import DjangoTestSuiteRunner

class PinDbTestSuiteRunner(DjangoTestSuiteRunner):
    # Hacking out db creation here so the test cases
    #  can handle it themselves.
    def setup_databases(self, **kwargs):
        pass

    def teardown_databases(self, old_config, **kwargs):
        pass



from pindb import populate_replicas
from django.utils.datastructures import SortedDict


DEBUG = True

TEST_RUNNER = "test_project.test_runner.PinDbTestSuiteRunner"

ROOT_URLCONF = 'test_project.urls'

INSTALLED_APPS = (
    'pindb',
    'test_app',
)

MIDDLEWARE_CLASSES = (
    'pindb.middleware.PinDbMiddleware',
)

SECRET_KEY = "1e1909jd10d9joKMLAKXklmax"


from __future__ import absolute_import

class PinDbException(Exception):
    pass

class PinDbConfigError(PinDbException, ValueError):
    pass


class UnpinnedWriteException(PinDbException):
    pass





from __future__ import absolute_import

from math import ceil
from time import time

from django.conf import settings

import anyjson

from . import pin, get_newly_pinned, unpin_all, is_enabled


# The name of the cookie that directs a request's reads to the master DB
PINNING_COOKIE = getattr(settings, 'PINDB_PINNING_COOKIE', 'pindb_pinned_set')

# The number of seconds for which reads are directed to the master DB after a
# write
PINNING_SECONDS = int(getattr(settings, 'PINDB_PINNING_SECONDS', 15))

def _get_request_pins(cookie_value):
    """Extract the persistent pinnings from a cookie.

    Return an iterable of (DB alias, time pinned until) tuples. Any expired
    pinnings are omitted.

    """
    ret = []

    now_time = time()
    try:
        pinned_untils = anyjson.loads(cookie_value)
    except ValueError:
        # If the cookie was corrupted, revert to not pinning anything:
        # TODO: Maybe add some logging.
        pinned_untils = []

    for alias, until in pinned_untils:
        if not alias in settings.MASTER_DATABASES:
            continue
        if now_time < until:
            ret.append((alias, until))
    return ret

def _get_response_pins(request_pinned_until):
    """Return the union of the preexisting pinned set--socked away on the request--with any newly set pins."""
    pinned_until = request_pinned_until.copy()

    # Update (a copy of) the previous persistent pinned set with any new pinnings:
    new_expiration = int(ceil(time() + PINNING_SECONDS))
    for alias in get_newly_pinned():
        pinned_until[alias] = new_expiration

    return pinned_until

class PinDbMiddleware(object):
    """Middleware to support the persisting pinning between requests after a write.

    Attaches a cookie to browser which has just written, causing subsequent
    DB reads (for some period of time, hopefully exceeding replication lag)
    to be handled by the master.

    When the cookie is detected on a request, related DBs are pinned.

    """
    def process_request(self, request):
        """Pin DB sets according to data in an incoming cookie."""
        # Make a clean slate. This is also necessary to ensure the threadlocal
        # attrs of our locals() object exist.
        unpin_all()

        request._pinned_until = {}

        if ((not is_enabled()) or 
                (not PINNING_COOKIE in request.COOKIES)):
            return

        for alias, until in _get_request_pins(request.COOKIES[PINNING_COOKIE]):
            # Keep track of existing end times for the return trip.
            request._pinned_until[alias] = until
            pin(alias, count_as_new=False)

    def process_response(self, request, response):
        """Set outgoing cookie to persist preexisting and new pinnings."""
        if not is_enabled():
            return response

        pinned_until = _get_response_pins(request._pinned_until)

        to_persist = list(pinned_until.items())
        # Don't set the cookie if there are no effective pins.
        if to_persist:
            # TODO: Use Django 1.4's signed cookies.
            response.set_cookie(PINNING_COOKIE,
                value=anyjson.dumps(to_persist),
                max_age=PINNING_SECONDS)

        return response



from __future__ import absolute_import

from copy import deepcopy
import os, tempfile
from threading import local

from django import VERSION as dj_VERSION
from django.http import HttpRequest, HttpResponse
from django import db as dj_db
from django.db.backends.dummy.base import DatabaseWrapper as DummyDatabaseWrapper
from django.db.utils import ConnectionHandler, ConnectionRouter
from django.conf import settings
from django.core.management import call_command
# TransactionTestCase is used instead of TestCase
#  because TestCase holds db changes in a pending transaction,
#  which are not visible from the replica connection (even though)
#  it is TEST_MIRROR'd.

from django.test import TransactionTestCase 
from django.test.simple import DjangoTestSuiteRunner
from django.utils import importlib

import anyjson
from mock import patch
from override_settings import override_settings  # a backport from Django 1.4

from test_project.test_app.models import HamModel, EggModel, FrobModel

import pindb
from pindb import middleware
from pindb.exceptions import PinDbConfigError, UnpinnedWriteException

"""
Test writing without pinning
Test writing after pinning
Test writing after unpinning
Test unmanaged aliases (not in MASTER/_SETS) work

Test round-robin DB
Test delegate router used
test that TEST_MIRROR was effective.
test NUM_REPLICAS set correctly.
Test web pinning context:
    test no cookie case
    test cookie with bad pin name (changed alias, user hacking)
    timeout
    persistence of different timestamps
    test wipes out dirty pins on request start
Test task context:
    test wipes out dirty pins on request start
    test write without pinning
    test writing after pinning
"""

class InitTestCase(TransactionTestCase):
    def setUp(self):
        super(InitTestCase, self).setUp()
        pindb._locals.pinned_set = set()
        pindb._locals.pinned_set.add('wark')
        pindb._locals.inited = False

    def test_init_worked(self):
        self.assertEqual(set(['wark']), pindb._locals.pinned_set)
        pindb._init_state()
        self.assertEqual(set(), pindb._locals.pinned_set)

    def test_init_happens_only_once(self):
        self.assertEqual(set(['wark']), pindb._locals.pinned_set)
        pindb._init_state()
        self.assertEqual(set(), pindb._locals.pinned_set)
        pindb._locals.pinned_set.add('wargle')
        pindb._init_state()
        self.assertEqual(set(['wargle']), pindb._locals.pinned_set)
        # except now we'll force inited back:
        pindb._locals.inited = False
        pindb._init_state()
        self.assertEqual(set(), pindb._locals.pinned_set)

# TODO: add coverage, COVERAGE_MODULE_EXCLUDES
class PinDbTestCase(TransactionTestCase):
    multi_db = True

    def _pre_setup(self):
        """Munge DB infrastructure before the superclass gets a chance to set up the DBs."""
        # clear all module state
        pindb._init_state()

        # patch up the db system to use the effective router settings (see override_settings)
        # we can't just reconstruct objects here because
        #   lots of places do from foo import baz, so they have
        #   a local reference to an object we can't replace.
        # so reach in and (gulp) mash the object's state.
        if dj_VERSION < (1, 4):
            for conn in dj_db.connections._connections.values():
                conn.close()
            dj_db.connections._connections = {}
        else:
            for conn in dj_db.connections.all():
                conn.close()
            dj_db.connections._connections = local()
        dj_db.connections.databases = settings.DATABASES

        def make_router(import_path):
            module_path, class_name = import_path.rsplit('.', 1)
            mod = importlib.import_module(module_path)
            return getattr(mod, class_name)()

        dj_db.router.routers = [
            make_router(import_path) for import_path in
            settings.DATABASE_ROUTERS]

        dj_db.connection = dj_db.connections[dj_db.DEFAULT_DB_ALIAS]
        dj_db.backend = dj_db.load_backend(dj_db.connection.settings_dict['ENGINE'])

        self.shim_runner = DjangoTestSuiteRunner()

        self.setup_databases()

        super(PinDbTestCase, self)._pre_setup()

    def _post_teardown(self):
        """Delete the databases after the superclass' method has closed the connections.

        We must do the DB deletion in post-teardown, because that's when the
        superclass closes its connection, which inadvertantly re-creates the
        sqlite file if we had previously tried to delete it (like in a plain
        old tearDown method).

        """
        super(PinDbTestCase, self)._post_teardown()
        pindb.unpin_all()
        self.teardown_databases(self.old_config)

    def setup_databases(self, **kwargs):
        self.old_config = self.shim_runner.setup_databases(**kwargs)

    def teardown_databases(self, old_config, **kwargs):
        self.shim_runner.teardown_databases(self.old_config)

    def _get_response_cookie(self, url):
        response = self.client.post(url)
        self.assertTrue(middleware.PINNING_COOKIE in response.cookies)
        return sorted(
                anyjson.loads(response.cookies[middleware.PINNING_COOKIE].value)
        )

def populate_databases(settings_dict):
    """Given a dict of various DB-related settings (DATABASES, MASTER_DATABASES, DATABASE_SETS), finalize all the settings into the DATABASES item of the dict.

    Also, turn the SQLite DBs we're using for testing into disk-based ones
    rather than memory-based ones, and give each a unique FS path, because
    Django buggily fails to come up with unique identifying tuples for multiple
    memory-based SQLites. Give each one a TEST_NAME, because otherwise,
    django.db.backends.sqlite3.creation reverts to ":memory:".

    """
    db_dir = tempfile.mkdtemp()
    if 'DATABASES' not in settings_dict:
        settings_dict['DATABASES'] = {}
    replicas =  pindb.populate_replicas(
        settings_dict['MASTER_DATABASES'],
        settings_dict['DATABASE_SETS']
    )

    settings_dict['DATABASES'].update(
        replicas
    )
    for alias, db in settings_dict['DATABASES'].items():
        db['NAME'] = db['TEST_NAME'] = os.path.join(db_dir, "test_%s" % alias)


misconfigured_settings = {
    'DATABASE_ROUTERS': ['pindb.StrictPinDbRouter'],
}  # because MASTER_DATAABASES and DATABASE_SETS is required.

@override_settings(**misconfigured_settings)
class MisconfiguredTest(TransactionTestCase):
    multi_db = True  # Necessary? Tests pass without.

    def test_router_catches_misconfiguration(self):
        self.assertRaises(PinDbConfigError, ConnectionRouter, settings.DATABASE_ROUTERS)

    def test_populate_replicas_catches_misconfiguration(self):
        no_default = {
            'DATABASE_ROUTERS': ['pindb.StrictPinDbRouter'],
            'MASTER_DATABASES': dict([
                ('ham', {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3',
                }),
                ('egg', {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3',
                })
            ]),
            'DATABASE_SETS': {
                'ham': [],
                # This normally would have overrides, but lots of sqlites in
                # memory are happy together:
                'egg': [{}, {}]
            },
            'PINDB_DELEGATE_ROUTERS': ['test_project.router.HamAndEggRouter']
        }

        self.assertRaises(PinDbConfigError,
            pindb.populate_replicas,
            no_default['MASTER_DATABASES'],
            {}
        )

        default_ok = no_default.copy()  # Note: not deep, but doesn't matter
        default_ok['MASTER_DATABASES']['default'] = default_ok['MASTER_DATABASES']['ham']
        del default_ok['MASTER_DATABASES']['ham']
        try:
            pindb.populate_replicas(
                default_ok['MASTER_DATABASES'],
                 {'default': [], 'egg': []})
        except PinDbConfigError:
            self.fail("Expected default to be acceptable config.")


no_delegate_router_settings = {
    'DATABASE_ROUTERS': ['pindb.StrictPinDbRouter'],
    'MASTER_DATABASES': {
        'default':  {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        },
        'egg': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'DATABASE_SETS': {
        # This normally would have overrides, but lots of sqlites in memory are
        # happy together:
        'default': [{}, {}],
        'egg': [] #no replicas
    },
    'PINDB_DELEGATE_ROUTERS': None
}
populate_databases(no_delegate_router_settings)

@override_settings(**no_delegate_router_settings)
class NoDelegateTest(PinDbTestCase):
    def test_internals(self):
        self.assertEqual(pindb.DB_SET_SIZES['default'], 1)
        self.assertEqual(pindb.DB_SET_SIZES['egg'], -1)

    def test_pinning(self):
        # pinning is reflected in is_pinned
        for master in settings.MASTER_DATABASES:
            self.assertFalse(pindb.is_pinned(master))
            pindb.pin(master)
            self.assertTrue(pindb.is_pinned(master))
        # unpin_all resets the state.
        pindb.unpin_all()
        for master in settings.MASTER_DATABASES:
            self.assertFalse(pindb.is_pinned(master))

        pindb.pin("default")
        # pinning state is available as a set.
        pinned1 = pindb.get_pinned()
        self.assertTrue('default' in pinned1)
        # unpinning doesn't affect the previous set.
        pindb.unpin_all()
        pinned2 = pindb.get_pinned()
        self.assertTrue('default' in pinned1)
        self.assertFalse('default' in pinned2)

        # we can keep track of new pins vs. carried-over pins
        #  i.e. pins that stick for the replication lag period.
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        pindb.pin("default")
        self.assertEqual(1, len(pindb.get_newly_pinned()))
        pindb.unpin_all()
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        pindb.pin("default", count_as_new=False)
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        # it counts as pinned even if we don't count it as new.
        self.assertEqual(1, len(pindb.get_pinned()))

    @patch("pindb.randint")
    def test_unpinned_replica(self, mock_randint):
        pindb.pin("default")
        self.assertEqual(
            dj_db.router.db_for_read(HamModel), "default"
        )
        with pindb.unpinned_replica("default"):
            mock_randint.return_value = 1
            self.assertEqual(
                dj_db.router.db_for_read(HamModel), "default-1"
            )

        with self.assertRaises(ValueError):
            with pindb.unpinned_replica("default"):
                raise ValueError

        @pindb.with_replicas("default")
        def to_replicas():
            mock_randint.return_value = 1
            self.assertEqual(
                dj_db.router.db_for_read(HamModel), "default-1"
            )
        to_replicas()

        @pindb.with_replicas(["default"])
        def to_replicas_list():
            mock_randint.return_value = 2
            self.assertEqual(
                dj_db.router.db_for_read(HamModel), "default-1"
            )
        to_replicas_list()

    @patch("pindb.randint")
    def test_with_master(self, mock_randint):
        mock_randint.return_value = 0
        self.assertEqual(
            dj_db.router.db_for_read(HamModel), "default-0"
        )
        with pindb.master("default"):
            self.assertEqual(
                dj_db.router.db_for_read(HamModel), "default"
            )
        self.assertEqual(
            dj_db.router.db_for_read(HamModel), "default-0"
        )

        @pindb.with_masters(["default"])
        def to_master():
            mock_randint.return_value = 2
            self.assertEqual(
                dj_db.router.db_for_read(HamModel), "default"
            )

    @patch("pindb.randint")
    def test_get_replica(self, mock_randint):
        mock_randint.return_value = 0
        self.assertEqual(pindb.get_replica("default"), "default-0")
        mock_randint.return_value = 1
        self.assertEqual(pindb.get_replica("default"), "default-0")

        # gets the master if there are no replicas
        self.assertEqual(pindb.get_replica("egg"), "egg")
        self.assertEqual(pindb.get_replica("frob"), "frob")

    def test_router(self):
        self.assertTrue(
            dj_db.router.db_for_read(HamModel) in ["default-0", "default-1"]
        )
        self.assertRaises(
            UnpinnedWriteException,
            dj_db.router.db_for_write,
            HamModel
        )
        self.assertTrue(
            dj_db.router.db_for_read(EggModel) in ["default-0", "default-1"]
        )
        self.assertRaises(
            UnpinnedWriteException,
            dj_db.router.db_for_write,
            EggModel
        )
        pindb.pin("default")
        try:
            self.assertEqual('default', dj_db.router.db_for_write(HamModel))
        except UnpinnedWriteException:
            self.fail("Expected to be able to write after pinning.")

        ham1 = HamModel.objects.create()
        ham2 = HamModel.objects.create()
        # If no delegate router is given, all DB goes to default.
        egg = EggModel.objects.create()
        self.assertTrue(dj_db.router.allow_relation(ham1, ham2))
        self.assertTrue(dj_db.router.allow_syncdb("default", HamModel))
        self.assertTrue(dj_db.router.allow_syncdb("default", EggModel))


delegate_strict_router_settings = {
    'DATABASE_ROUTERS': ['pindb.StrictPinDbRouter'],
    'DATABASES': {
        'frob': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'MASTER_DATABASES': {
        'default': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        },
        'egg': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'DATABASE_SETS': {
        # This normally would have overrides, but lots of sqlites in memory are
        # happy together:
        'default': [],
        'egg': [{}, {}]
    },
    'PINDB_DELEGATE_ROUTERS': ['test_project.router.HamAndEggRouter']
}
populate_databases(delegate_strict_router_settings)

@override_settings(**delegate_strict_router_settings)
class FullyConfiguredStrictTest(PinDbTestCase):
    def test_internals(self):
        self.assertEqual(pindb.DB_SET_SIZES['default'], -1)
        self.assertEqual(pindb.DB_SET_SIZES['egg'], 1)

    def test_pinning(self):
        # pinning is reflected in is_pinned
        for master in settings.MASTER_DATABASES:
            self.assertFalse(pindb.is_pinned(master))
            pindb.pin(master)
            self.assertTrue(pindb.is_pinned(master))
        # unpin_all resets the state.
        pindb.unpin_all()
        for master in settings.MASTER_DATABASES:
            self.assertFalse(pindb.is_pinned(master))

        pindb.pin("default")
        # pinning state is available as a set.
        pinned1 = pindb.get_pinned()
        self.assertTrue('default' in pinned1)
        # unpinning doesn't affect the previous set.
        pindb.unpin_all()
        pinned2 = pindb.get_pinned()
        self.assertTrue('default' in pinned1)
        self.assertFalse('default' in pinned2)

        # we can keep track of new pins vs. carried-over pins
        #  i.e. pins that stick for the replication lag period.
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        pindb.pin("default")
        self.assertEqual(1, len(pindb.get_newly_pinned()))
        pindb.unpin_all()
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        pindb.pin("default", count_as_new=False)
        self.assertEqual(0, len(pindb.get_newly_pinned()))
        # it counts as pinned even if we don't count it as new.
        self.assertEqual(1, len(pindb.get_pinned()))

    @patch("pindb.randint")
    def test_get_replica(self, mock_randint):
        mock_randint.return_value = 0
        self.assertEqual(pindb.get_replica("default"), "default")
        mock_randint.return_value = 1
        self.assertEqual(pindb.get_replica("default"), "default")

        # gets the master if there are no replicas
        mock_randint.return_value = 0
        self.assertEqual(pindb.get_replica("egg"), "egg-0")
        mock_randint.return_value = 1
        self.assertEqual(pindb.get_replica("egg"), "egg-0")

        # nonexistent or unmanaged DATABASES should return the alias
        self.assertEqual(pindb.get_replica("frob"), "frob")
        self.assertEqual(pindb.get_replica("nope"), "nope")

    def test_router(self):
        self.assertEqual(
            dj_db.router.db_for_read(HamModel), "default"
        )
        self.assertRaises(
            UnpinnedWriteException,
            dj_db.router.db_for_write,
            HamModel
        )
        self.assertTrue(
            dj_db.router.db_for_read(EggModel) in ["egg-0", "egg-1"]
        )
        self.assertRaises(
            UnpinnedWriteException,
            dj_db.router.db_for_write,
            EggModel
        )
        pindb.pin("default")
        try:
            self.assertEqual('default', dj_db.router.db_for_write(HamModel))
        except UnpinnedWriteException:
            self.fail("Expected to be able to write after pinning.")

        ham1 = HamModel.objects.create()
        ham2 = HamModel.objects.create()
        # pinning a different DB doesn't allow writes on others.
        self.assertRaises(
            UnpinnedWriteException,
            dj_db.router.db_for_write,
            EggModel
        )

        pindb.pin("egg")
        try:
            egg = EggModel.objects.create()
        except UnpinnedWriteException:
            self.fail("Expected to be able to write after pinning.")

        self.assertTrue(dj_db.router.allow_relation(ham1, ham2))
        self.assertTrue(dj_db.router.allow_syncdb("default", HamModel))
        self.assertTrue(dj_db.router.allow_syncdb("default", EggModel))

        self.assertEqual(dj_db.router.db_for_read(FrobModel), "default")


delegate_greedy_router_settings = {
    'DATABASE_ROUTERS': ['pindb.GreedyPinDbRouter'],
    'DATABASES': {
        'frob': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'MASTER_DATABASES': {
        'default': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        },
        'egg': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'DATABASE_SETS': {
        # This normally would have overrides, but lots of sqlites in memory are
        # happy together:
        'default': [],
        'egg': [{}, {}]
    },
    'PINDB_DELEGATE_ROUTERS': ['test_project.router.HamAndEggRouter']
}
greedy_middleware_settings = deepcopy(delegate_greedy_router_settings)
populate_databases(delegate_greedy_router_settings)
populate_databases(greedy_middleware_settings)  # for GreedyMiddlewareTest

@override_settings(**delegate_greedy_router_settings)
class FullyConfiguredGreedyTest(PinDbTestCase):
    """Tests for a ``GreedyPinDbRouter``, complete with delegate router"""

    def test_router(self):
        # Ham should go in the default DB:
        self.assertEqual(dj_db.router.db_for_read(HamModel), "default")
        # Reading one shouldn't cause it to pin:
        self.assertEqual(pindb.is_pinned("default"), False)
        # The delegate router has no opinion on Hams, so they should write to
        # the default DB:
        self.assertEqual(dj_db.router.db_for_write(HamModel), "default")
        # The above should cause it to pin (even though default has no slaves
        # in this case):
        self.assertEqual(pindb.is_pinned("default"), True)

        # Eggs are stored in a replicated DB set, so we should read from a replica:
        self.assertTrue(dj_db.router.db_for_read(EggModel) in ["egg-0", "egg-1"])
        # Reading shouldn't pin:
        self.assertEqual(pindb.is_pinned("egg"), False)
        # Writes should go to master:
        self.assertEqual(dj_db.router.db_for_write(EggModel), "egg")
        # And should cause a pin:
        self.assertEqual(pindb.is_pinned("egg"), True)
        pindb.unpin_all()

        # Making a Ham should pin to the master of the default set...
        ham1 = HamModel.objects.create()
        self.assertEqual(pindb.is_pinned("default"), True)
        # ...but not the egg set:
        self.assertEqual(pindb.is_pinned("egg"), False)

        # Making an Egg should pin the egg set as well:
        egg1 = EggModel.objects.create()
        self.assertEqual(pindb.is_pinned("egg"), True)
        pindb.unpin_all()

        ham1a = HamModel.objects.get(pk=ham1.pk)
        egg1a = EggModel.objects.get(pk=egg1.pk)  # no longer bewm

    def test_new_pins_persist(self):
        """If a greedy router scoops up a new pinning, make sure it counts as new.

        If it doesn't middleware and such will fail to persist it.

        """
        # Unprovoked by a write, it returns a replica:
        self.assertTrue(dj_db.router.db_for_read(EggModel) in ["egg-0", "egg-1"])
        # Then we write:
        self.assertEqual(dj_db.router.db_for_write(EggModel), "egg")
        # And then it should be in the newly pinned set:
        self.assertTrue(pindb.is_newly_pinned("egg"))

    def test_misdirected_save(self):
        """Test that saves route to the right DB after a fetch.

        At some point in production, we had an error that looked like fetching
        a model instance (from a slave) and then saving it failed, because it
        was trying to to write to the slave. It sure doesn't seem to to that
        [anymore].

        """
        # Set up the pre-existing model instance:
        egg = EggModel.objects.create()
        egg_id = egg.id
        pindb.unpin_all()

        egg = EggModel.objects.get(id=egg_id)
        self.assertEqual(dj_db.router.db_for_write(EggModel), "egg")


@override_settings(**greedy_middleware_settings)
class GreedyMiddlewareTest(PinDbTestCase):
    def test_read(self):
        response = self.client.post('/test_app/read/')
        self.assertFalse(middleware.PINNING_COOKIE in response.cookies)

    @patch('pindb.middleware.time')
    def test_write(self, mock_time):
        mock_time.return_value = 1
        cookie = self._get_response_cookie('/test_app/write/')
        self.assertEqual(cookie, [["default", 1 + middleware.PINNING_SECONDS]])
        mock_time.return_value = 2
        cookie = self._get_response_cookie('/test_app/write/')
        self.assertEqual(cookie, [["default", 2 + middleware.PINNING_SECONDS]])

    @patch('pindb.middleware.time')
    def test_write_with_existing(self, mock_time):
        mock_time.return_value = 1
        cookie = self._get_response_cookie('/test_app/write/')
        self.assertEqual(cookie, [["default", 1 + middleware.PINNING_SECONDS]])
        mock_time.return_value = 2

        cookie = self._get_response_cookie('/test_app/create_one_pin/')
        self.assertEqual(cookie, [
            ["default", 1 + middleware.PINNING_SECONDS],
            ["egg", 2 + middleware.PINNING_SECONDS],
        ])

    def test_bad_cookie(self):
        self.assertEquals(middleware._get_request_pins('bad thing'), [])

disabled_settings = {
    'PINDB_ENABLED': False,
    'DATABASE_ROUTERS': ['pindb.GreedyPinDbRouter'],
    'MASTER_DATABASES': {
        'default':  {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        },
        'egg': {
            'NAME': ':memory:',
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'DATABASE_SETS': {
        # This normally would have overrides, but lots of sqlites in memory are
        # happy together:
        'default': [{}, {}],
        'egg': [] #no replicas
    },
    'PINDB_DELEGATE_ROUTERS': ['test_project.router.HamAndEggRouter']
}
populate_databases(disabled_settings)

@override_settings(**disabled_settings)
class DisabledTest(PinDbTestCase):
    def test_read(self):
        response = self.client.post('/test_app/read/')
        self.assertFalse(middleware.PINNING_COOKIE in response.cookies)
    
    def test_write(self):
        response = self.client.post('/test_app/write/')
        self.assertFalse(middleware.PINNING_COOKIE in response.cookies)

    def test_router(self):
        # Ham should go in the default DB:
        self.assertEqual(dj_db.router.db_for_read(HamModel), "default")
        # Reading one shouldn't cause it to pin:
        self.assertEqual(pindb.is_pinned("default"), False)
        # The delegate router has no opinion on Hams, so they should write to
        # the default DB:
        self.assertEqual(dj_db.router.db_for_write(HamModel), "default")
        # The above should *not* cause it to pin since we're disabled.
        self.assertEqual(pindb.is_pinned("default"), False)

        # Eggs are stored in a replicated DB set, but since we're disabled, 
        #   we should still get the master.
        self.assertEqual(dj_db.router.db_for_read(EggModel), "egg")
        # Reading shouldn't pin:
        self.assertEqual(pindb.is_pinned("egg"), False)
        # Writes should go to master:
        self.assertEqual(dj_db.router.db_for_write(EggModel), "egg")
        # but still should *not* cause it to pin since we're disabled.
        self.assertEqual(pindb.is_pinned("egg"), False)



from django.http import HttpResponse
from .models import HamModel, EggModel
import pindb

def create_no_pins(request):
    with pindb.master("default"):
        HamModel.objects.create()
    return HttpResponse("create no pins")

def create_one_pin(request):
    with pindb.master("default"):
        HamModel.objects.create()
    EggModel.objects.create()
    return HttpResponse("create one pin")

def write(request):
    HamModel.objects.create()
    return HttpResponse("write")

def read(request):
    HamModel.objects.all()
    return HttpResponse("read")



from django.conf.urls.defaults import *
from . import views

urlpatterns = patterns('',
    # fixme: {% url %} instead of hard-code in tests.
    url(r"read", views.read, name='read'),
    url(r"write", views.write, name='write'),
    url(r"create_no_pins", views.create_no_pins, name='create_no_pins'),
    url(r"create_one_pin", views.create_one_pin, name='create_one_pin'),
)



from .test_app.models import HamModel, EggModel

class HamAndEggRouter(object):
    """Read and write Eggs to the 'egg' DB. Ham goes in the default DB."""
    def db_for_read(self, model, **hints):        
        if model is EggModel:
            return "egg"

    def db_for_write(self, model, **hints):
        if model is EggModel:
            return "egg"

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_syncdb(self, db, model):
        if db == "egg" and model is EggModel:
            return True
        if db == "default" and model is HamModel:
            return True
        return None



#!/usr/bin/env python

import os
import sys
import codecs

try:
    from setuptools import setup, Command, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, Command, find_packages  # noqa
from distutils.command.install import INSTALL_SCHEMES

src_dir = "pindb"

class RunTests(Command):
    description = "Run the django test suite from the tests dir."
    user_options = []
    extra_env = {}
    extra_args = []

    def run(self):
        for env_name, env_value in self.extra_env.items():
            os.environ[env_name] = str(env_value)

        this_dir = os.getcwd()
        testproj_dir = os.path.join(this_dir, "test_project")
        os.chdir(testproj_dir)
        sys.path.append(testproj_dir)
        from django.core.management import execute_manager
        os.environ["DJANGO_SETTINGS_MODULE"] = os.environ.get(
                        "DJANGO_SETTINGS_MODULE", "settings")
        settings_file = os.environ["DJANGO_SETTINGS_MODULE"]
        settings_mod = __import__(settings_file, {}, {}, [''])
        prev_argv = list(sys.argv)
        try:
            sys.argv = [__file__, "test"] + self.extra_args
            execute_manager(settings_mod, argv=sys.argv)
        finally:
            sys.argv = prev_argv

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


if os.path.exists("README.rst"):
    long_description = codecs.open("README.rst", "r", "utf-8").read()
else:
    long_description = "See https://github.com/votizen/django-pindb"

setup(
    name = 'django-pindb',
    packages=find_packages(),
    version='0.1.12', # remember to change __init__
    description = 'Manages master/replica pinning for django',
    long_description = long_description,
    url = 'https://github.com/votizen/django-pindb',
    author = 'Jeremy Dunck',
    author_email = 'jdunck@votizen.com',
    maintainer = 'Jeremy Dunck',
    maintainer_email = 'jdunck@votizen.com',
    keywords = ['django', 'multidb', 'router'],
    license = 'MIT',
    install_requires=['anyjson'],
    tests_require=['django>=1.2.0', 'django-override-settings>=1.2', 'mock>=0.7.2'],
    cmdclass={
        "test": RunTests,
    },
    include_package_data=True,
    platforms=["any"],
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
